from mockmpi import mock_mpiexec


def f(comm, txt, index=0, index2=0):
    assert txt == "abc"
    assert index == index2


def test_args():
    # direct args
    mock_mpiexec(2, f, "abc")
    mock_mpiexec(2, f, "abc", index="cat", index2="cat")

    # if have args as list and kwargs as dict, can dereference them.
    args = ["abc"]
    mock_mpiexec(2, f, *args)
    kwargs = {"index": "cat", "index2": "cat"}
    mock_mpiexec(2, f, *args, **kwargs)

    # can also dereference as literals, which is parobably weird, but allowed.
    mock_mpiexec(2, f, *["abc"], **{"index": "cat", "index2": "cat"})

if __name__ == '__main__':
    test_args()



# Copyright (c) Mike Jarvis and other collaborators
# See https://github.com/rmjarvis/MockMPI/LICENSE for license information.

from ._version import __version__, __version_info__
from .comm import MockComm
from .exec import mock_mpiexec



# Copyright (c) Mike Jarvis and other collaborators
# See https://github.com/rmjarvis/MockMPI/LICENSE for license information.

__version__ = '0.8.2'
__version_info__ = tuple(map(int, __version__.split('.')))




# Copyright (c) Mike Jarvis and other collaborators
# See https://github.com/rmjarvis/MockMPI/LICENSE for license information.

import numpy as np

# This constant used to be 1 in both MPICH and OpenMPI,
# but starting with mpi4py version 4, they switched it to -1.
# Use -1 here, but when we check for it allow 1 as well.
# And if we happen to have mpi4py installed, include whatever it actually has as well.
try:
    from mpi4py.MPI import IN_PLACE
except ImportError:
    IN_PLACE = -1
ALLOWED_IN_PLACE = [IN_PLACE, 1, -1]


class MockComm(object):
    """A class to mock up the MPI Comm API using a multiprocessing Pipe.

    """

    def __init__(self, rank, size, pipes, barrier):
        self.rank = rank
        self.size = size
        self.pipes = pipes
        self.barrier = barrier

    def __bool__(self):
        return self.size > 0

    def Get_rank(self):
        return self.rank

    def Get_size(self):
        return self.size

    def send(self, msg, dest):
        if dest != self.rank:
            self.pipes[dest].send(msg)
        else:
            self.msg = msg

    def Send(self, msg, dest):
        if not isinstance(msg, np.ndarray):
            raise ValueError(
                "Can only use Send with numpy arrays "
                "(Mocking code does not handle general buffers)"
            )
        self.send(msg, dest)

    def recv(self, source):
        if source != self.rank:
            msg = self.pipes[source].recv()
        else:
            msg = self.msg
        return msg

    def Recv(self, buffer, source):
        msg = self.recv(source)
        buffer[:] = msg

    def Barrier(self):
        self.barrier.wait()

    def bcast(self, msg, root=0):
        if root == self.rank:
            for p in range(self.size):
                self.send(msg, p)
        msg = self.recv(root)
        return msg

    def Bcast(self, msg, root=0):
        if root == self.rank:
            for p in range(self.size):
                self.Send(msg, p)
        self.Recv(msg, root)

    def scatter(self, data, root=0):
        if root == self.rank:
            for p in range(self.size):
                self.send(data[p], p)
        data = self.recv(root)
        return data

    def gather(self, data, root=0):
        self.send(data, root)
        if root == self.rank:
            new_data = []
            for p in range(self.size):
                new_data.append(self.recv(p))
            return new_data
        else:
            return None

    def alltoall(self, data=0):
        for p in range(self.size):
            self.send(data[p], p)
        new_data = []
        for p in range(self.size):
            new_data.append(self.recv(p))
        return new_data

    def reduce(self, sendobj, op=None, root=0):
        if op is not None:
            raise NotImplementedError("Not implemented non-sum reductions in mock MPI")
        new_data = self.gather(sendobj, root)

        if root == self.rank:
            d = new_data[0]
            for d2 in new_data[1:]:
                d = d + d2
            return d
        else:
            return None

    def allreduce(self, sendobj, op=None):
        d = self.reduce(sendobj, op)
        d = self.bcast(d)
        return d

    def Reduce(self, sendbuf, recvbuf, op=None, root=0):
        if isinstance(sendbuf, int) and (sendbuf in ALLOWED_IN_PLACE):
            sendbuf = recvbuf.copy()

        if not isinstance(sendbuf, np.ndarray):
            raise ValueError(
                "Cannot use Reduce with non-arrays. "
                "(Mocking code does not handle general buffers)"
            )

        r = self.reduce(sendbuf, op=op, root=root)
        if self.rank == root:
            recvbuf[:] = r

    def Allreduce(self, sendbuf, recvbuf, op=None):
        self.Reduce(sendbuf, recvbuf, op)
        self.Bcast(recvbuf)

    def allgather(self, sendobj):
        obj = self.gather(sendobj)
        return self.bcast(obj)

    # Instance methods not implemented
    def Abort(self, *args, **kwargs):
        raise NotImplementedError("The method 'Abort' is not implemented in mockmpi")

    def Accept(self, *args, **kwargs):
        raise NotImplementedError("The method 'Accept' is not implemented in mockmpi")

    def Allgather(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Allgather' is not implemented in mockmpi"
        )

    def Allgatherv(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Allgatherv' is not implemented in mockmpi"
        )

    def Alltoall(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Alltoall' is not implemented in mockmpi"
        )

    def Alltoallv(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Alltoallv' is not implemented in mockmpi"
        )

    def Alltoallw(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Alltoallw' is not implemented in mockmpi"
        )

    def Bsend(self, *args, **kwargs):
        raise NotImplementedError("The method 'Bsend' is not implemented in mockmpi")

    def Bsend_init(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Bsend_init' is not implemented in mockmpi"
        )

    def Call_errhandler(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Call_errhandler' is not implemented in mockmpi"
        )

    def Cart_map(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Cart_map' is not implemented in mockmpi"
        )

    def Clone(self, *args, **kwargs):
        raise NotImplementedError("The method 'Clone' is not implemented in mockmpi")

    def Connect(self, *args, **kwargs):
        raise NotImplementedError("The method 'Connect' is not implemented in mockmpi")

    def Create(self, *args, **kwargs):
        raise NotImplementedError("The method 'Create' is not implemented in mockmpi")

    def Create_cart(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Create_cart' is not implemented in mockmpi"
        )

    def Create_dist_graph(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Create_dist_graph' is not implemented in mockmpi"
        )

    def Create_dist_graph_adjacent(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Create_dist_graph_adjacent' is not implemented in mockmpi"
        )

    def Create_graph(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Create_graph' is not implemented in mockmpi"
        )

    def Create_group(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Create_group' is not implemented in mockmpi"
        )

    def Create_intercomm(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Create_intercomm' is not implemented in mockmpi"
        )

    def Delete_attr(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Delete_attr' is not implemented in mockmpi"
        )

    def Disconnect(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Disconnect' is not implemented in mockmpi"
        )

    def Dup(self, *args, **kwargs):
        raise NotImplementedError("The method 'Dup' is not implemented in mockmpi")

    def Dup_with_info(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Dup_with_info' is not implemented in mockmpi"
        )

    def Exscan(self, *args, **kwargs):
        raise NotImplementedError("The method 'Exscan' is not implemented in mockmpi")

    def Free(self, *args, **kwargs):
        raise NotImplementedError("The method 'Free' is not implemented in mockmpi")

    def Gather(self, *args, **kwargs):
        raise NotImplementedError("The method 'Gather' is not implemented in mockmpi")

    def Gatherv(self, *args, **kwargs):
        raise NotImplementedError("The method 'Gatherv' is not implemented in mockmpi")

    def Get_attr(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Get_attr' is not implemented in mockmpi"
        )

    def Get_errhandler(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Get_errhandler' is not implemented in mockmpi"
        )

    def Get_group(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Get_group' is not implemented in mockmpi"
        )

    def Get_info(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Get_info' is not implemented in mockmpi"
        )

    def Get_name(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Get_name' is not implemented in mockmpi"
        )

    def Get_topology(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Get_topology' is not implemented in mockmpi"
        )

    def Graph_map(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Graph_map' is not implemented in mockmpi"
        )

    def Iallgather(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Iallgather' is not implemented in mockmpi"
        )

    def Iallgatherv(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Iallgatherv' is not implemented in mockmpi"
        )

    def Iallreduce(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Iallreduce' is not implemented in mockmpi"
        )

    def Ialltoall(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Ialltoall' is not implemented in mockmpi"
        )

    def Ialltoallv(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Ialltoallv' is not implemented in mockmpi"
        )

    def Ialltoallw(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Ialltoallw' is not implemented in mockmpi"
        )

    def Ibarrier(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Ibarrier' is not implemented in mockmpi"
        )

    def Ibcast(self, *args, **kwargs):
        raise NotImplementedError("The method 'Ibcast' is not implemented in mockmpi")

    def Ibsend(self, *args, **kwargs):
        raise NotImplementedError("The method 'Ibsend' is not implemented in mockmpi")

    def Idup(self, *args, **kwargs):
        raise NotImplementedError("The method 'Idup' is not implemented in mockmpi")

    def Iexscan(self, *args, **kwargs):
        raise NotImplementedError("The method 'Iexscan' is not implemented in mockmpi")

    def Igather(self, *args, **kwargs):
        raise NotImplementedError("The method 'Igather' is not implemented in mockmpi")

    def Igatherv(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Igatherv' is not implemented in mockmpi"
        )

    def Improbe(self, *args, **kwargs):
        raise NotImplementedError("The method 'Improbe' is not implemented in mockmpi")

    def Iprobe(self, *args, **kwargs):
        raise NotImplementedError("The method 'Iprobe' is not implemented in mockmpi")

    def Irecv(self, *args, **kwargs):
        raise NotImplementedError("The method 'Irecv' is not implemented in mockmpi")

    def Ireduce(self, *args, **kwargs):
        raise NotImplementedError("The method 'Ireduce' is not implemented in mockmpi")

    def Ireduce_scatter(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Ireduce_scatter' is not implemented in mockmpi"
        )

    def Ireduce_scatter_block(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Ireduce_scatter_block' is not implemented in mockmpi"
        )

    def Irsend(self, *args, **kwargs):
        raise NotImplementedError("The method 'Irsend' is not implemented in mockmpi")

    def Is_inter(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Is_inter' is not implemented in mockmpi"
        )

    def Is_intra(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Is_intra' is not implemented in mockmpi"
        )

    def Iscan(self, *args, **kwargs):
        raise NotImplementedError("The method 'Iscan' is not implemented in mockmpi")

    def Iscatter(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Iscatter' is not implemented in mockmpi"
        )

    def Iscatterv(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Iscatterv' is not implemented in mockmpi"
        )

    def Isend(self, *args, **kwargs):
        raise NotImplementedError("The method 'Isend' is not implemented in mockmpi")

    def Issend(self, *args, **kwargs):
        raise NotImplementedError("The method 'Issend' is not implemented in mockmpi")

    def Mprobe(self, *args, **kwargs):
        raise NotImplementedError("The method 'Mprobe' is not implemented in mockmpi")

    def Probe(self, *args, **kwargs):
        raise NotImplementedError("The method 'Probe' is not implemented in mockmpi")

    def Recv_init(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Recv_init' is not implemented in mockmpi"
        )

    def Reduce_scatter(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Reduce_scatter' is not implemented in mockmpi"
        )

    def Reduce_scatter_block(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Reduce_scatter_block' is not implemented in mockmpi"
        )

    def Rsend(self, *args, **kwargs):
        raise NotImplementedError("The method 'Rsend' is not implemented in mockmpi")

    def Rsend_init(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Rsend_init' is not implemented in mockmpi"
        )

    def Scan(self, *args, **kwargs):
        raise NotImplementedError("The method 'Scan' is not implemented in mockmpi")

    def Scatter(self, *args, **kwargs):
        raise NotImplementedError("The method 'Scatter' is not implemented in mockmpi")

    def Scatterv(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Scatterv' is not implemented in mockmpi"
        )

    def Send_init(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Send_init' is not implemented in mockmpi"
        )

    def Sendrecv(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Sendrecv' is not implemented in mockmpi"
        )

    def Sendrecv_replace(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Sendrecv_replace' is not implemented in mockmpi"
        )

    def Set_attr(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Set_attr' is not implemented in mockmpi"
        )

    def Set_errhandler(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Set_errhandler' is not implemented in mockmpi"
        )

    def Set_info(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Set_info' is not implemented in mockmpi"
        )

    def Set_name(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Set_name' is not implemented in mockmpi"
        )

    def Spawn(self, *args, **kwargs):
        raise NotImplementedError("The method 'Spawn' is not implemented in mockmpi")

    def Spawn_multiple(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Spawn_multiple' is not implemented in mockmpi"
        )

    def Split(self, *args, **kwargs):
        raise NotImplementedError("The method 'Split' is not implemented in mockmpi")

    def Split_type(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Split_type' is not implemented in mockmpi"
        )

    def Ssend(self, *args, **kwargs):
        raise NotImplementedError("The method 'Ssend' is not implemented in mockmpi")

    def Ssend_init(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'Ssend_init' is not implemented in mockmpi"
        )

    def barrier(self, *args, **kwargs):
        raise NotImplementedError("The method 'barrier' is not implemented in mockmpi")

    def bsend(self, *args, **kwargs):
        raise NotImplementedError("The method 'bsend' is not implemented in mockmpi")

    def exscan(self, *args, **kwargs):
        raise NotImplementedError("The method 'exscan' is not implemented in mockmpi")

    def ibsend(self, *args, **kwargs):
        raise NotImplementedError("The method 'ibsend' is not implemented in mockmpi")

    def improbe(self, *args, **kwargs):
        raise NotImplementedError("The method 'improbe' is not implemented in mockmpi")

    def iprobe(self, *args, **kwargs):
        raise NotImplementedError("The method 'iprobe' is not implemented in mockmpi")

    def irecv(self, *args, **kwargs):
        raise NotImplementedError("The method 'irecv' is not implemented in mockmpi")

    def isend(self, *args, **kwargs):
        raise NotImplementedError("The method 'isend' is not implemented in mockmpi")

    def issend(self, *args, **kwargs):
        raise NotImplementedError("The method 'issend' is not implemented in mockmpi")

    def mprobe(self, *args, **kwargs):
        raise NotImplementedError("The method 'mprobe' is not implemented in mockmpi")

    def probe(self, *args, **kwargs):
        raise NotImplementedError("The method 'probe' is not implemented in mockmpi")

    def py2f(self, *args, **kwargs):
        raise NotImplementedError("The method 'py2f' is not implemented in mockmpi")

    def scan(self, *args, **kwargs):
        raise NotImplementedError("The method 'scan' is not implemented in mockmpi")

    def sendrecv(self, *args, **kwargs):
        raise NotImplementedError(
            "The method 'sendrecv' is not implemented in mockmpi"
        )

    def ssend(self, *args, **kwargs):
        raise NotImplementedError("The method 'ssend' is not implemented in mockmpi")

    # Properties not implemented

    @property
    def topology(self):
        raise NotImplementedError(
            "The property 'topology' is not implemented in mockmpi"
        )

    @property
    def group(self):
        raise NotImplementedError("The property 'group' is not implemented in mockmpi")

    @property
    def name(self):
        raise NotImplementedError("The property 'name' is not implemented in mockmpi")

    @property
    def is_inter(self):
        raise NotImplementedError(
            "The property 'is_inter' is not implemented in mockmpi"
        )

    @property
    def is_intra(self):
        raise NotImplementedError(
            "The property 'is_intra' is not implemented in mockmpi"
        )

    @property
    def is_topo(self):
        raise NotImplementedError(
            "The property 'is_topo' is not implemented in mockmpi"
        )

    # 'Info' is the only writeable property
    @property
    def info(self):
        raise NotImplementedError("The property 'info' is not implemented in mockmpi")

    @info.setter
    def info(self, *args, **kwargs):
        raise NotImplementedError("The property 'info' is not implemented in mockmpi")

    # Class methods not yet implemented
    @classmethod
    def Compare(cls, *args, **kwargs):
        raise NotImplementedError(
            "The class method 'Compare' is not implemented in mockmpi"
        )

    @classmethod
    def Get_parent(cls, *args, **kwargs):
        raise NotImplementedError(
            "The class method 'Get_parent' is not implemented in mockmpi"
        )

    @classmethod
    def Join(cls, *args, **kwargs):
        raise NotImplementedError(
            "The class method 'Join' is not implemented in mockmpi"
        )

    @classmethod
    def Create_keyval(cls, *args, **kwargs):
        raise NotImplementedError(
            "The class method 'Create_keyval' is not implemented in mockmpi"
        )

    @classmethod
    def Free_keyval(cls, *args, **kwargs):
        raise NotImplementedError(
            "The class method 'Free_keyval' is not implemented in mockmpi"
        )

    @classmethod
    def f2py(cls, *args, **kwargs):
        raise NotImplementedError(
            "The class method 'f2py' is not implemented in mockmpi"
        )



# Copyright (c) Mike Jarvis and other collaborators
# See https://github.com/rmjarvis/MockMPI/LICENSE for license information.

import numpy as np
import multiprocessing as mp
from .comm import MockComm

# We use this subclass to help with exception
# handling, as described here:
# https://stackoverflow.com/a/33599967/989692
class Process(mp.Process):
    def __init__(self, *args, **kwargs):
        mp.Process.__init__(self, *args, **kwargs)
        self._pconn, self._cconn = mp.Pipe()
        self._exception = None

    def run(self):
        try:
            mp.Process.run(self)
            self._cconn.send(None)
        except Exception as e:
            self._cconn.send(e)
            raise e

    @property
    def exception(self):
        if self._pconn.poll():
            self._exception = self._pconn.recv()
        return self._exception


def mock_mpiexec(nproc, target, *args, **kwargs):
    """Run a function, given as target, as though it were an MPI session using mpiexec -n nproc
    but using multiprocessing instead of mpi.
    """
    mp.set_start_method("spawn", force=True)

    # Make the message passing pipes
    all_pipes = [{} for p in range(nproc)]
    for i in range(nproc):
        for j in range(i + 1, nproc):
            p1, p2 = mp.Pipe()
            all_pipes[i][j] = p1
            all_pipes[j][i] = p2

    # Make a barrier
    barrier = mp.Barrier(nproc)

    # Make fake MPI-like comm object
    comms = [
        MockComm(rank, nproc, pipes, barrier) for rank, pipes in enumerate(all_pipes)
    ]

    # Make processes
    procs = [Process(target=target, args=(comm,) + args, kwargs=kwargs) for comm in comms]

    for p in procs:
        p.start()

    for p in procs:
        d = p.join()
        if p.exception:
            raise p.exception.__class__ from p.exception



# Copyright (c) Mike Jarvis and other collaborators
# See https://github.com/rmjarvis/MockMPI/LICENSE for license information.

from mockmpi import mock_mpiexec
import numpy as np

def run_mpi_session(comm):
    """A simple MPI session we want to run in mock MPI mode.

    This serves as a test of comm.py and exec.py
    """
    rank = comm.Get_rank()
    size = comm.Get_size()
    print(rank, "rank, size = ", rank, size, flush=True)

    comm.send("Hello! My rank is %d." % rank, dest=0)
    print(rank, "sent hello message ", flush=True)

    if rank == 0:
        for p in range(size):
            print(rank, "try to read from ", p, flush=True)
            msg = comm.recv(source=p)
            print(rank, "received message: ", repr(msg), flush=True)

    print(rank, "Before barrier", flush=True)
    comm.Barrier()
    print(rank, "After barrier", flush=True)

    if rank == 0:
        data = np.arange(size) + 10
    else:
        data = None

    print(rank, "Before bcast: data = ", data, flush=True)
    data = comm.bcast(data, root=0)
    print(rank, "After bcast: data = ", data, flush=True)
    np.testing.assert_array_equal(data, np.arange(size) + 10)
    comm.Barrier()

    if rank == 0:
        data = np.arange(size) + 10
    else:
        data = np.empty(size, dtype=int)

    print(rank, "Before Bcast: data = ", data, flush=True)
    comm.Bcast(data, root=0)
    print(rank, "After Bcast: data = ", data, flush=True)
    np.testing.assert_array_equal(data, np.arange(size) + 10)
    comm.Barrier()

    if rank != 0:
        data = None

    print(rank, "Before scatter: data = ", data, flush=True)
    data = comm.scatter(data, root=0)
    print(rank, "After scatter: data = ", data, flush=True)
    assert data == rank + 10
    comm.Barrier()

    print(rank, "Before gather: data = ", data, flush=True)
    data = comm.gather(data, root=0)
    print(rank, "After gather: data = ", data, flush=True)
    if rank == 0:
        np.testing.assert_array_equal(data, np.arange(size) + 10)
    else:
        assert data is None
    comm.Barrier()

    data = np.arange(size) + rank ** 2 + 5
    print(rank, "Before alltoall: data = ", data, flush=True)
    data = comm.alltoall(data)
    print(rank, "After alltoall: data = ", data, flush=True)
    np.testing.assert_array_equal(data, np.arange(size) ** 2 + rank + 5)

    # test reduction
    if size < 52:
        letters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
        x = letters[rank]
        s = comm.reduce(x)
        if rank == 0:
            assert s == letters[:size]
        else:
            assert s is None
        comm.Barrier()
    else:
        print("Skipping reduction test - too large")

    # test all reduction
    if size < 52:
        letters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
        x = letters[rank]
        s = comm.allreduce(x)
        assert s == letters[:size]
        comm.Barrier()
    else:
        print("Skipping reduction test - too large")

    # test Reduce
    x = np.zeros(10, dtype=float) + rank
    y = np.zeros(10, dtype=float)
    print(rank, "Before Reduce: x = ", x, flush=True)
    print(rank, "Before Reduce: y = ", y, flush=True)
    comm.Reduce(x, y)
    if rank == 0:
        assert np.allclose(y, size * (size - 1) // 2)
    print(rank, "After Reduce: x = ", x, flush=True)
    print(rank, "After Reduce: y = ", y, flush=True)

    # # test All Reduce
    x = np.zeros(10, dtype=float) + rank
    y = np.zeros(10, dtype=float)
    print(rank, "Before AllReduce: x = ", x, flush=True)
    print(rank, "Before AllReduce: y = ", y, flush=True)
    comm.Allreduce(x, y)
    print(rank, "After AllReduce: x = ", x, flush=True)
    print(rank, "After AllReduce: y = ", y, flush=True)
    assert np.allclose(y, size * (size - 1) // 2)

    # test allgather
    my_item = rank * 7
    all_items = comm.allgather(my_item)
    print(rank, "after allgather, all_items = ", all_items)
    assert len(all_items) == size
    assert all_items[rank] == my_item
    for i in range(size):
        assert all_items[i] == i * 7


def test_mpi_session():
    mock_mpiexec(2, run_mpi_session)
    mock_mpiexec(4, run_mpi_session)


if __name__ == "__main__":
    # Test this code.
    test_mpi_session()



from setuptools import setup, find_packages
import pathlib
import re

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

# Only uses numpy and multiprocessing
dependencies = ['numpy']

# Read in the version from mockmpi/_version.py
# cf. http://stackoverflow.com/questions/458550/standard-way-to-embed-version-into-python-package
version_file = (here / 'mockmpi' / '_version.py')
verstrline = version_file.read_text(encoding='utf-8')
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
version = mo.group(1)
print('mockmpi version is %s'%(version))

setup(
    name='MockMPI',
    version=version,
    description='A tool for mocking mpi4py for testing',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/rmjarvis/MockMPI',
    download_url='https://github.com/rmjarvis/MockMPI/releases/tag/v%s.zip'%version,
    author='Mike Jarvis and collaborators',
    author_email='michael@jarvis.net',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3 :: Only',
    ],
    keywords='MPI, development, unittesting',
    packages=['mockmpi'],
    python_requires='>=3.5',
    install_requires=dependencies,
)
